from django.apps import apps
from django.db.models import Q
from django.shortcuts import _get_queryset

from jackal.exceptions import NotFound
from jackal.loaders import structure_loader
from jackal.settings import jackal_settings


def get_object_or_None(klass, *args, **kwargs):
    queryset = _get_queryset(klass)
    try:
        return queryset.get(*args, **kwargs)
    except queryset.model.DoesNotExist:
        return None


def get_object_or(klass, this=None, *args, **kwargs):
    return get_object_or_None(klass, *args, **kwargs) or this


def get_object_or_404(model, **fields):
    obj = get_object_or_None(model, **fields)
    if obj is None:
        raise NotFound(model, fields)
    return obj


def model_update(instance, **fields):
    for key, value in fields.items():
        setattr(instance, key, value)

    instance.save()
    return instance


def get_model(*args, **kwargs):
    return apps.get_model(*args, **kwargs)


def operating(a, oper, b):
    if oper == '==':
        return a == b
    elif oper == '<=':
        return a <= b
    elif oper == '<':
        return a < b
    elif oper == '>=':
        return a >= b
    elif oper == '>':
        return a > b
    elif oper == '!=':
        return a != b
    else:
        return False


def status_checker(change_to, current, key):
    stru = structure_loader('STATUS_CONDITION_CLASSES').get(key)
    if stru is None:
        return False

    condition = stru.get(change_to)

    if condition is None:
        return False

    for operator, target_status in condition:
        if not operating(current, operator, target_status):
            return False

    return True


def status_readable(status, key):
    stru = structure_loader('STATUS_READABLE_CLASSES').get(key, {})
    return stru.get(status, jackal_settings.UNKNOWN_READABLE)


def gen_q(key, *filter_keywords):
    q_object = Q()
    for q in filter_keywords:
        q_object |= Q(**{q: key})
    return q_object



class JackalBaseStructure:
    prefix = 'structure'

    @classmethod
    def get_structure(cls):
        """
        정의된 prefix 값을 이용하여 해당 클래스 내의 {prefix}_ 형태로 시작하는 모든 항목들을 가져옵니다.
        만약 항목이 함수라면, 해당 함수를 실행한 값을 최종 딕셔너리에 업데이트하고, 일반 딕셔너리라면 바로 업데이트합니다.
        사전도, 함수도 아니라면 건너 뜁니다.
        """
        ret_dict = {}
        for key in cls.__dict__.keys():
            if key.find('{}_'.format(cls.prefix)) == 0:
                value = getattr(cls, key, {})
                if callable(value):
                    ret_dict.update(value())
                elif type(value) is dict:
                    ret_dict.update(value)
                else:
                    continue

        return ret_dict


class BaseStatusCondition(JackalBaseStructure):
    pass


class BaseStatusReadable(JackalBaseStructure):
    pass


class BaseQueryFunction:
    prefix = 'func'

    @classmethod
    def get_function_set(cls):
        ret_dict = {}

        for name, func in cls.__dict__.items():
            if name.find('{}_'.format(cls.prefix)) == 0:
                name = name.replace('{}_'.format(cls.prefix), '')
                ret_dict[name] = func.__func__
        return ret_dict


class DefaultQueryFunction(BaseQueryFunction):
    @staticmethod
    def func_to_list(data):
        return data.split(',')

    @staticmethod
    def func_to_boolean(data):
        return data.lower() == 'true'



import datetime
from calendar import monthrange
from datetime import timedelta

from dateutil.relativedelta import relativedelta
from django.utils import timezone


def local_date(days=0, months=0, years=0, **kwargs):
    """
    This function use TIMEZONE of django settings.

    when today is 2018-01-02 # YYYY-MM-DD

    >>> local_date()
    => datetime.date(2018, 1, 2)
    >>> local_date(days=1)
    => datetime.date(2018, 1, 3)
    >>> local_date(days=-1)
    => datetime.date(2017, 1, 1)
    >>> local_date(months=-1)
    => datetime.date(2018, 12, 2)
    >>> local_date(months=1)
    => datetime.date(2018, 2, 2)
    >>> local_date(day=1)
    => datetime.date(2018, 1, 1)
    >>> local_date(day=-1)
    => datetime.date(2018, 1, 31)
    >>> local_date(day=1, months=1)
    => datetime.date(2018, 2, 1)
    >>> local_date(day=-1, months=1)
    => datetime.date(2018, 2, 28)
    >>> local_date(year=2019, month=5, day=18)
    => datetime.date(2019, 5, 18)
    >>> local_date(day=-1, months=1, days=1)
    => datetime.date(2018, 3, 1)
    """
    today = timezone.localdate()

    if kwargs.get('day') == -1:
        kwargs['day'] = get_last_day_of_month(
            today.year if kwargs.get('year') is None else kwargs.get('year'),
            today.month if kwargs.get('month') is None else kwargs.get('month')
        )

    today = today.replace(**kwargs)
    return today + relativedelta(days=days, months=months, years=years)


def local_tomorrow(today=None, skip_weekend=False):
    """
    This function almost same as local_date. But it return tomorrow of given today (or local today)
    """
    if today is None:
        today = local_date()

    add_days = 1

    if skip_weekend is True:
        if today.weekday() >= 4:
            add_days = 7 - today.weekday()

    return local_date(days=add_days, month=today.month, day=today.day, year=today.year)


def local_time(hours=0, minutes=0, seconds=0, microseconds=0, to_time=False, **kwargs):
    now = timezone.localtime().replace(**kwargs)
    now += timedelta(minutes=minutes, hours=hours, seconds=seconds, microseconds=microseconds)
    if to_time:
        return now.time()
    return now


def date_range(start, end):
    if type(start) is str:
        start = datetime.datetime.strptime(start, '%Y-%m-%d')
    if type(end) is str:
        end = datetime.datetime.strptime(end, '%Y-%m-%d')

    step = timedelta(days=1)
    ret_list = []
    while start <= end:
        ret_list.append(start)
        start += step
    return ret_list


def get_last_day_of_month(year, month):
    return monthrange(year, month)[1]



from rest_framework import serializers

from jackal.paginators import JackalPaginator
from jackal.shortcuts import model_update
from jackal.views.base import JackalAPIView


class ListCreateGeneric(JackalAPIView):
    def list(self, request, **kwargs):
        serializer_class = self.get_serializer_class()
        filtered_queryset = self.get_filtered_queryset(request, **kwargs)
        ser = serializer_class(instance=filtered_queryset, many=True)
        return self.simple_response(ser.data)

    def create(self, request, **kwargs):
        model = self.get_model()
        obj = model.objects.create(**request.inspected_data)
        return self.success(id=obj.id)


class DetailUpdateDestroyGeneric(JackalAPIView):
    def detail(self, request, **kwargs):
        obj = self.get_object(request, **kwargs)
        ser = self.serializer_class(obj)
        return self.simple_response(ser.data)

    def update(self, request, **kwargs):
        obj = self.get_object(request, **kwargs)
        model_update(obj, **request.inspected_data)
        return self.success(id=obj.id)

    def destroy(self, request, **kwargs):
        obj = self.get_object(request, **kwargs)
        obj.delete()
        return self.success()


class PaginateListGeneric(JackalAPIView):
    default_page_number = 1
    default_page_length = None
    page_number_key = 'page_number'
    page_length_key = 'page_length'

    def paginated_list(self, request, **kwargs):
        queryset = self.get_filtered_queryset(request, **kwargs)
        return self.paginated_data(request, queryset)

    def paginated_data(self, request, queryset):
        page_num = request.query_params.get(self.page_number_key, self.default_page_number)
        page_length = request.query_params.get(self.page_length_key, self.default_page_length)
        paginator = JackalPaginator(queryset, page_num, page_length)
        response_data = paginator.serialized_data(
            self.get_serializer_class(), self.get_serializer_context()
        )
        return self.simple_response(response_data)


class SimpleGeneric(JackalAPIView):
    def wrapper(self, request, method='get', **kwargs):
        return getattr(self, '{}_func'.format(method))(request, **kwargs)

    def get_func(self, request, **kwargs):
        pass

    def post_func(self, request, **kwargs):
        pass

    def patch_func(self, request, **kwargs):
        pass

    def delete_func(self, request, **kwargs):
        pass


class LabelValueListGeneric(JackalAPIView):
    label_field = 'name'
    value_field = 'id'

    def get_serializer_class(self):
        class LabelValueSerializer(serializers.ModelSerializer):
            label = serializers.CharField(source=self.label_field)
            value = serializers.CharField(source=self.value_field)

            class Meta:
                model = self.get_model()
                fields = ('label', 'value')

        return LabelValueSerializer

    def label_value_list(self, request, **kwargs):
        queryset = self.get_filtered_queryset(request, **kwargs)
        ser_class = self.get_serializer_class()
        return self.simple_response(ser_class(queryset, many=True))


class DetailMixin(DetailUpdateDestroyGeneric):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)


class DestroyMixin(DetailUpdateDestroyGeneric):
    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class UpdateMixin(DetailUpdateDestroyGeneric):
    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)


class DetailUpdateMixin(DetailUpdateDestroyGeneric):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)

    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)


class DetailUpdateDestroyMixin(DetailUpdateDestroyGeneric):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)

    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)

    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class DetailDestroyMixin(DetailUpdateDestroyGeneric):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)

    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class UpdateDestroyMixin(DetailUpdateDestroyGeneric):
    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)

    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class ListMixin(ListCreateGeneric):
    def get(self, request, **kwargs):
        return self.list(request, **kwargs)


class CreateMixin(ListCreateGeneric):
    def post(self, request, **kwargs):
        return self.create(request, **kwargs)


class ListCreateMixin(ListCreateGeneric):
    def get(self, request, **kwargs):
        return self.list(request, **kwargs)

    def post(self, request, **kwargs):
        return self.create(request, **kwargs)


class PaginateListMixin(PaginateListGeneric):
    def get(self, request, **kwargs):
        return self.paginated_list(request, **kwargs)


class PostMixin(SimpleGeneric):
    def post(self, request, **kwargs):
        return self.wrapper(request, method='post', **kwargs)


class GetMixin(SimpleGeneric):
    def get(self, request, **kwargs):
        return self.wrapper(request, method='get', **kwargs)


class PutMixin(SimpleGeneric):
    def put(self, request, **kwargs):
        return self.wrapper(request, method='put', **kwargs)


class PatchMixin(SimpleGeneric):
    def patch(self, request, **kwargs):
        return self.wrapper(request, method='patch', **kwargs)


class DeleteMixin(SimpleGeneric):
    def delete(self, request, **kwargs):
        return self.wrapper(request, method='delete', **kwargs)


class LabelValueListMixin(LabelValueListGeneric):
    def get(self, request, **kwargs):
        return self.label_value_list(request, **kwargs)



import collections

import six


class JackalDictMapper:
    @staticmethod
    def av2bv(a_dict, b_dict):
        return {a_value: b_dict.get(a_key) for a_key, a_value in a_dict.items()}


def isiter(arg):
    return isinstance(arg, collections.Iterable) and not isinstance(arg, six.string_types)
